#ifndef __ABAC_H__
#define __ABAC_H__

#include <stdlib.h>
#include <stdbool.h>
#include <abac_common.h>

typedef struct _abac_context_t abac_context_t;

typedef struct _abac_credential_t abac_credential_t;
typedef struct _abac_aspect_t abac_aspect_t;
typedef struct _abac_attribute_t abac_attribute_t;
typedef struct _abac_id_t abac_id_t;
typedef struct _abac_id_credential_t abac_id_credential_t;

typedef struct _abac_condition_t abac_condition_t;
typedef struct _abac_term_t abac_term_t;
typedef struct _abac_item_t abac_item_t;
typedef struct _abac_param_list_t abac_param_list_t;
typedef struct _abac_key_t abac_key_t;

/* called within abac.hh */
char *abac_version();
void abac_errx(int,const char *);
abac_condition_t *abac_condition_dup(abac_condition_t *ptr);
void abac_condition_free(abac_condition_t *ptr);
char *abac_constraint_string(abac_condition_t *ptr);
char *abac_constraint_typed_string(abac_condition_t *ptr);
int file_exist(char *filename);

void abac_terms_free(abac_term_t **terms);
abac_term_t *abac_term_dup(abac_term_t *ptr);
void abac_term_free(abac_term_t *);
abac_term_t *abac_term_named_create(char*);
abac_term_t *abac_term_create(char*, char*, abac_condition_t*);
abac_term_t *abac_term_id_create(abac_id_t *ptr);
abac_term_t *abac_term_dup(abac_term_t *ptr);
char *abac_term_typed_string(abac_term_t *ptr);
char *abac_term_string(abac_term_t *ptr);
int abac_term_is_time(abac_term_t *);
int abac_term_is_urn(abac_term_t *term);
int abac_term_is_integer(abac_term_t *term);
int abac_term_is_string(abac_term_t *term);
abac_term_t *abac_term_add_constraint(abac_term_t *ptr, abac_condition_t *cond);
int abac_term_type(abac_term_t *term);
char *abac_term_name(abac_term_t *);

void abac_aspects_free(abac_aspect_t **aspects);
abac_aspect_t *abac_aspect_dup(abac_aspect_t *ptr);
void abac_aspect_free(abac_aspect_t *);
char *abac_aspect_string(abac_aspect_t *ptr);
char *abac_aspect_typed_string(abac_aspect_t *ptr);

abac_param_list_t *abac_aspect_aspect_params(abac_aspect_t *ptr);
abac_term_t **abac_param_list_vectorize(abac_param_list_t *ptr);
abac_param_list_t *abac_aspect_linked_role_params(abac_aspect_t *ptr);
char *abac_oset_string(abac_aspect_t *ptr);
char *abac_oset_typed_string(abac_aspect_t *ptr);
char *abac_role_string(abac_aspect_t *ptr);
char *abac_role_typed_string(abac_aspect_t *ptr);

void abac_id_print_key_chunk(abac_id_t *id);
char *abac_id_cert_filename(abac_id_t *id);
char *abac_id_privkey_filename(abac_id_t *id);
char *abac_id_subject(abac_id_t *id);
char *abac_id_issuer(abac_id_t *id);
abac_id_t *abac_id_dup(abac_id_t *id);
abac_id_t *abac_id_credential_id(abac_id_credential_t *ptr);
abac_id_t *abac_id_from_file(char *filename);
abac_id_t *abac_id_from_chunk(abac_chunk_t);
int abac_id_generate(abac_id_t **ret, char *cn, int validity);
int abac_id_generate_with_key(abac_id_t **ret, char *cn, int validity, char *keyfile, char *pfile);
char *abac_id_keyid(abac_id_t *id);
char *abac_id_cn(abac_id_t *id);
void abac_id_free(abac_id_t *id);
int abac_id_has_privkey(abac_id_t *id);
void abac_id_write_cert(abac_id_t *id, FILE *out);
int abac_id_write_privkey(abac_id_t *id, FILE *out);
abac_chunk_t abac_id_cert_chunk(abac_id_t *id);
abac_chunk_t abac_id_privkey_chunk(abac_id_t *id);
char *abac_id_string(abac_id_t *id);
int abac_id_load_privkey_file(abac_id_t *id, char *filename);
void abac_id_write_cert_fname(abac_id_t *id, char *filename);
void abac_id_write_privkey_fname(abac_id_t *id, char *filename);
int abac_id_load_encrypted_privkey_file(abac_id_t *ptr, char *filename, 
char *pfile);

abac_key_t *abac_key_generate();

abac_attribute_t *abac_attribute_dup(abac_attribute_t *ptr);
void abac_attribute_free(abac_attribute_t *ptr);
abac_attribute_t *abac_credential_attribute(abac_credential_t *cred);
int abac_attribute_create(abac_attribute_t **ret,abac_aspect_t *head, abac_aspect_t *tail,int validity);
abac_attribute_t *abac_attribute_add_tail(abac_attribute_t *ptr, abac_aspect_t *);
char *abac_attribute_string(abac_attribute_t *ptr);
char *abac_attribute_typed_string(abac_attribute_t *ptr);
abac_attribute_t *abac_attribute_from_chunk(abac_chunk_t achunk);

abac_aspect_t *abac_attribute_head(abac_attribute_t *ptr);
abac_aspect_t *abac_attribute_tail(abac_attribute_t *ptr);

int abac_attribute_is_role(abac_attribute_t *ptr);
abac_aspect_t **abac_attribute_tail_vectorized(abac_attribute_t *ptr);
int abac_attribute_bake(abac_attribute_t *ptr);
int abac_attribute_baked(abac_attribute_t *ptr);
int abac_attribute_write_cert(abac_attribute_t *ptr, FILE *out);
abac_chunk_t abac_attribute_cert_chunk(abac_attribute_t *ptr);
void abac_attribute_write_cert_fname(abac_attribute_t *,char*);

abac_id_credential_t *abac_context_principal_lookup(abac_context_t *ctx, char *name);
void abac_context_principals_free(abac_id_credential_t **principals);
abac_context_t *abac_context_new(void);
abac_context_t *abac_context_dup(abac_context_t *ctx);
void abac_context_free(abac_context_t *ctx);
int abac_context_load_encrypted_id_file(abac_context_t *ptr, char *filename,
char *pfile);
int abac_context_load_encrypted_id_files(abac_context_t *ptr, char *filename,
char *keyfilename, char *pfile);
int abac_context_load_id_enc_privkey_chunk(abac_context_t *ctx,
abac_chunk_t cert, abac_chunk_t privkey, char *pfile); 

void show_yap_db(const char *msg);

abac_credential_t **abac_context_query(abac_context_t *ctx, char *role, char *principal, int *success);
abac_credential_t **abac_context_query_again(abac_context_t *ctx, int *success);
abac_credential_t **abac_context_query_with_structure(abac_context_t *, abac_aspect_t*, abac_aspect_t*, int*);
void abac_context_credentials_free(abac_credential_t **);
abac_credential_t **abac_context_credentials(abac_context_t *);
abac_id_credential_t **abac_context_principals(abac_context_t *);

/*
 * Error codes for loading certificates.
 */
#define ABAC_CERT_SUCCESS           0   // certificate loaded, all is well
#define ABAC_CERT_INVALID           -1  // invalid format; also file not found
#define ABAC_CERT_BAD_SIG           -2  // invalid signature
#define ABAC_CERT_MISSING_ISSUER    -3  // missing ID cert that issued the attribute cert
#define ABAC_CERT_BAD_CN            -4  // ID cert is not matching CN=principal format
#define ABAC_CERT_BAD_YAP           -5  // failed to insert into prolog engine
#define ABAC_CERT_EXISTS            1   // ID already exists (does not default to a failure)

#define ABAC_ID_SUCCESS                     0
#define ABAC_ID_GENERATE_INVALID_CN        -1
#define ABAC_ID_GENERATE_INVALID_VALIDITY  -2

#define ABAC_ATTRIBUTE_SUCCESS              0
#define ABAC_ATTRIBUTE_INVALID_ROLE        -1
#define ABAC_ATTRIBUTE_INVALID_VALIDITY    -2
#define ABAC_ATTRIBUTE_ISSUER_NOKEY        -3
#define ABAC_ATTRIBUTE_FAIL                -4

#define ABAC_TERM_SUCCESS                   0
#define ABAC_TERM_FAIL                     -1

#define USE(evalue) ((getenv(evalue)!=NULL)?1:0)

/* from abac_c.c, the c interface */
void abac_constraint_add_integer_max(abac_condition_t *cond,int val);
void abac_constraint_add_integer_min(abac_condition_t *cond,int val);
void abac_constraint_add_integer_target(abac_condition_t *cond,int val);
void abac_constraint_add_float_max(abac_condition_t *cond,float val);
void abac_constraint_add_float_min(abac_condition_t *cond,float val);
void abac_constraint_add_float_target(abac_condition_t *cond,float val);
void abac_constraint_add_time_max(abac_condition_t *cond,char* val);
void abac_constraint_add_time_min(abac_condition_t *cond,char* val);
void abac_constraint_add_time_target(abac_condition_t *cond,char* val);
void abac_constraint_add_urn_target(abac_condition_t *cond,char* val);
void abac_constraint_add_string_target(abac_condition_t *cond,char* val);
void abac_constraint_add_boolean_target(abac_condition_t *cond, char* val);
abac_condition_t *abac_constraint_from_role(abac_aspect_t *ptr);
abac_condition_t *abac_constraint_from_oset(abac_aspect_t *ptr);
abac_condition_t *abac_constraint_create(char *vtype);

abac_aspect_t *abac_role_principal_create(char *principal_name);
abac_aspect_t *abac_oset_principal_create(char *principal_name);
abac_aspect_t *abac_oset_create(char *principal_name, char *oset_name);
abac_aspect_t *abac_role_create(char *principal_name, char *role_name);
abac_aspect_t *abac_oset_linked_create(char *principal_name, char* linked_role_name, char *oset_name);
abac_aspect_t *abac_role_linked_create(char *principal_name, char* linked_role_name, char *role_name);
abac_aspect_t *abac_oset_object_create(abac_term_t *object);
abac_aspect_t *abac_oset_create(char *principal_name, char *oset_name);
int abac_oset_is_object(abac_aspect_t *ptr);
int abac_oset_is_principal(abac_aspect_t *ptr);
int abac_role_is_principal(abac_aspect_t *ptr);
int abac_oset_is_linking(abac_aspect_t *ptr);
int abac_role_is_linking(abac_aspect_t *ptr);
char *abac_role_linked_role(abac_aspect_t *ptr);
char *abac_oset_linked_role(abac_aspect_t *ptr);
char *abac_role_name(abac_aspect_t *ptr);
char *abac_oset_name(abac_aspect_t *ptr);
char *abac_role_principal(abac_aspect_t *ptr);
char *abac_oset_principal(abac_aspect_t *ptr);
void abac_role_add_data_term(abac_aspect_t *ptr, abac_term_t *param);
void abac_oset_add_data_term(abac_aspect_t *ptr, abac_term_t *param);
void abac_role_add_linked_data_term(abac_aspect_t *ptr, abac_term_t *param);
void abac_oset_add_linked_data_term(abac_aspect_t *ptr, abac_term_t *param);
char *abac_oset_object(abac_aspect_t *ptr);
abac_term_t **abac_role_get_linked_data_terms(abac_aspect_t *ptr);
abac_term_t **abac_role_get_data_terms(abac_aspect_t *ptr);
abac_term_t **abac_oset_get_linked_data_terms(abac_aspect_t *ptr);
abac_term_t **abac_oset_get_data_terms(abac_aspect_t *ptr);
abac_aspect_t **abac_attribute_role_tails(abac_attribute_t *ptr);
abac_aspect_t **abac_attribute_oset_tails(abac_attribute_t *ptr);


char *abac_head_string(abac_attribute_t *ptr);
char *abac_tail_string(abac_attribute_t *ptr);
char *abac_head_typed_string(abac_attribute_t *ptr);
char *abac_tail_typed_string(abac_attribute_t *ptr);

int abac_context_load_id_chunk(abac_context_t *ctx, abac_chunk_t cert);
int abac_context_load_id_privkey_chunk(abac_context_t *ctx, abac_chunk_t cert, abac_chunk_t privkey);
int abac_context_load_attribute_file(abac_context_t *ctx, char *filename);
int abac_context_load_id(abac_context_t *ptr, abac_id_t *id);
int abac_context_load_id_file(abac_context_t *ptr, char *filename);
int abac_context_load_id_files(abac_context_t *ptr, char *filename, char *keyfilename);
int abac_context_load_attribute(abac_context_t *ptr, abac_attribute_t *attr);
void abac_context_load_directory(abac_context_t *ctx, char *path);
void abac_context_load_principals(abac_context_t *ctx, char *path);
int abac_context_load_attribute_chunk(abac_context_t *ctx, abac_chunk_t cert);
void abac_context_set_no_partial_proof(abac_context_t *ctx);
void abac_context_set_want_partial_proof(abac_context_t *ctx);

int file_exist(char *filename);

#endif /* __ABAC_H__ */
